"likfit.new" <-
  function (geodata, coords=geodata$coords, data=geodata$data, trend = "cte",
            ini.cov.pars,
            fix.nugget = FALSE, nugget = 0, 
            fix.kappa = TRUE, kappa = 0.5, 
            fix.lambda = TRUE, lambda = 1, 
            fix.anR = TRUE, anR = 1, 
            fix.anA = TRUE, anA = 0, 
            cov.model = c("matern", "exponential", "gaussian", "spherical",
              "circular", "cubic", "wave", "powered.exponential",
              "cauchy", "gneiting", "gneiting.matern", "pure.nugget"),
            method = "ML",
            components = FALSE, nospatial = TRUE,
            range.limits, func.inv = "chol",
            messages.screen = TRUE, ...) 
{
  ##
  ## Checking input
  ##
  if(is.R()) require(mva)
  call <- match.call()
  temp.list <- list()
  ##
  tausq <- nugget
  cov.model <- match.arg(cov.model)
  if (method == "REML" | method == "reml" | method == "rml") 
    method <- "RML"
  if(method == "ML" | method == "ml")
    method <- "ML"
  temp.list$method <- method
  if(length(ini.cov.pars) != 2)
    stop("ini.cov.pars must be a vector with 2 components: \ninitial values for sigmasq and phi")
  if(fix.kappa & !is.null(kappa))
    if(cov.model == "matern" & kappa == 0.5)
      cov.model <- "exponential"
  coords <- as.matrix(coords)
  n <- temp.list$n <- length(data)
  if ((2*n) != length(coords))
    stop("Number of locations does not match with number of data")
  temp.list$xmat <- trend.spatial(trend=trend, coords=coords)
  beta.size <- temp.list$beta.size <- dim(temp.list$xmat)[2]  
  ##
  ## Box-Cox transformation for fixed lambda
  ##
  if(fix.lambda) {
    if(round(lambda, dig=4) == 1) {
      temp.list$log.jacobian <- 0
      temp.list$z <- as.vector(data)
    }
    else {
      if(any(data <= 0))
        stop("Transformation option not allowed when there are zeros or negative data")
      Jdata <- data^(lambda - 1)
      if(any(Jdata <= 0))
        temp.list$log.jacobian <- log(prod(Jdata))
      else temp.list$log.jacobian <- sum(log(Jdata))
      Jdata <- NULL
      if(round(lambda, dig=4) == 0)
        temp.list$z <- log(data)
      else temp.list$z <- ((data^lambda) - 1)/lambda
    }
  }
  else{
    temp.list$z <- as.vector(data)
    temp.list$log.jacobian <- NULL
  }
  ##
  ## Coordinates transformation for fixed anisotropy parameters
  ##
  if(fix.anR & fix.anA){
    if(anR != 1 | anA != 0)
      coords <- coords.aniso(coords, aniso.pars=c(anA, anR))
    .likGRF.dists.vec <<- as.vector(dist(coords))
    range.dist <- range(.likGRF.dists.vec)
    max.dist <- max(range.dist)
    min.dist <- min(range.dist)
  }
  else temp.list$coords <- coords
  temp.list$cov.model <- cov.model
  temp.list$func.inv <- func.inv
  if(missing(range.limits)){
    lower.phi <- 0
    upper.phi <- +Inf
  }
  else{
    lower.phi <- range.limits[1]
    upper.phi <- range.limits[2]
  }
  ##
  ini <- ini.cov.pars[2]
##  fixed.pars <- NULL
  lower.optim <- c(lower.phi)
  upper.optim <- c(upper.phi)
  fixed.values <- list()
  ##
  if(fix.nugget) {
##    fixed.pars <- c(fixed.pars, 0)
    fixed.values$tausq <- nugget
  }
  else {
    ini <- c(ini, nugget/ini.cov.pars[1])
    lower.optim <- c(lower.optim, 0)
    upper.optim <- c(upper.optim, +Inf)
  }
  if(fix.kappa){
##    fixed.kappa <- c(fixed.pars, kappa)
    fixed.values$kappa <- kappa
  }
  else {
    ini <- c(ini, kappa)
    lower.optim <- c(lower.optim, 0)
    upper.optim <- c(upper.optim, +Inf)
  }
  if(fix.lambda){
##    fixed.pars <- c(fixed.pars, lambda)
    fixed.values$lambda <- lambda
  }
  else {
    ini <- c(ini, lambda)
    lower.optim <- c(lower.optim, -3)
    upper.optim <- c(upper.optim, 3)
  }
  if(fix.anR){
##    fixed.pars <- c(fixed.pars, anR)
    fixed.values$anR <- anR
  }
  else {
    ini <- c(ini, anR)
    lower.optim <- c(lower.optim, 1)
    upper.optim <- c(upper.optim, +Inf)
  }
  if(fix.anA){
##    fixed.pars <- c(fixed.pars, anA)
    fixed.values$anA <- anA
  }
  else {
    ini <- c(ini, anA)
    lower.optim <- c(lower.optim, 0)
    upper.optim <- c(upper.optim, 2*pi)
  }
  ## This must be here, after the previous ones:
  if(fix.nugget & nugget > 0){
    ## Warning: Inverting order here, ini will be now: c(phi,sigmasg)
    ini <- c(ini, ini.cov.pars[1])
    lower.optim <- c(lower.optim, 0)
    upper.optim <- c(upper.optim, +Inf)
##    fixed.pars <- c(fixed.pars, ini.cov.pars[1])
##    fixed.values$sigmasq <- 0
  }
  names(ini) <- NULL
  ##
  ip <- list(f.tausq = fix.nugget, f.kappa = fix.kappa, f.lambda = fix.lambda,
                  f.anR = fix.anR, f.anA = fix.anA)
  ##  
  if (messages.screen == TRUE) {
    cat(paste("likfit: Initialising likelihood maximisation using the function optim\n"))
    cat("--------------------------------------------------------------------\n")
    cat("likfit: consider providing control() arguments for optim.\n        For more details see documentation for the function optim.\n")
    cat("likfit: it is highly advisable to run this function several\n        times with different initial values for the parameters.\n")
    cat("--------------------------------------------------------------------\n")
    cat(paste("likfit: range of values allowed for the parameter:", nice(lower.phi), "to", nice(upper.phi), "\n"))
    cat("likfit: WARNING: This step can be time demanding!\n")
  }
  npars <- beta.size + 2 + sum(unlist(ip)==FALSE)
  if(is.R())
    lik.optim <- optim(par = ini, fn = negloglik.GRF, method="L-BFGS-B",
                       lower=lower.optim, upper=upper.optim,
                       fp=fixed.values, ip=ip, temp.list = temp.list, ...)
  else
    lik.results <- nlminb(ini, negloglik.GRF,
                          lower=lower.optim, upper=upper.optim,
                          fp=fixed.values, ip=ip, temp.list = temp.list, ...)
  ##
  if(messages.screen == TRUE) 
    cat("likfit: end of numerical maximisation.\n")
  par.est <- lik.optim$par
  phi <- par.est[1]
  ##
  ## Values of the maximised likelihood
  ##
  if(is.R())
    value.min <- lik.optim$value
  else
    value.min <- lik.optim$objective
  if(method == "ML"){
    if(ip$f.tausq & (tausq > 0))
      loglik.max <-  (- value.min) + (n/2)*(-log(2*pi))
    else
      loglik.max <-  (- value.min) + (n/2)*(-log(2*pi) + log(n) -1)
  }
  if(method == "RML"){
    xx.eigen <- eigen(crossprod(temp.list$xmat), symmetric = TRUE, only.values = TRUE)
    if(ip$f.tausq & (tausq > 0))
      loglik.max <- (- value.min) + ((n-beta.size)/2)*(-log(2*pi)) +
        0.5 * sum(log(xx.eigen$values))
    else
      loglik.max <- (- value.min) + ((n-beta.size)/2)*(-log(2*pi)) +
        ((n-beta.size)/2)*(log(n-beta.size)) - (n/2) + 0.5 * sum(log(xx.eigen$values))
  }
  ##
  ## Assigning values for estimated parameters
  ##
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    anA <- par.est[2]
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    anR <- par.est[2]
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    anR <- par.est[2]
    anA <- par.est[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    lambda  <- par.est[2]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    lambda  <- par.est[2]
    anA <- par.est[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    lambda  <- par.est[2]
    anR <- par.est[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    lambda  <- par.est[2]
    anR <- par.est[3]
    anA <- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    kappa  <-  par.est[2]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    kappa  <-  par.est[2]
    anA <- par.est[3]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    kappa  <-  par.est[2]
    anR <- par.est[3]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    kappa  <-  par.est[2]
    anR <- par.est[3]
    anA <- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
    anA <- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
    anR<- par.est[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    kappa <-  par.est[2]
    lambda <- par.est[3]
    anR<- par.est[4]
    anA<- par.est[5]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    anA<- par.est[3]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    anR<- par.est[3]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    anR<- par.est[3]
    anA<- par.est[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    lambda <- par.est[3]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    lambda <- par.est[3]
    anA <- par.est[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    lambda <- par.est[3]
    anR <- par.est[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    lambda <- par.est[3]
    anR <- par.est[4]
    anA <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    anA <- par.est[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    anR <- par.est[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    anR <- par.est[4]
    anA <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
    anA <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
    anR <- par.est[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- par.est[2]
    kappa <-  par.est[3]
    lambda <- par.est[4]
    anR <- par.est[5]
    anA <- par.est[6]
  }
  ##
  if(fix.nugget & nugget > 0){
    sigmasq <- par.est[length(par.est)]
    if(sigmasq > 1e-12) tausq <- nugget/sigmasq
    check.sigmasq <- TRUE
  }
  else check.sigmasq <- FALSE
  ##
  ## Transforming data acccording to the estimated lambda (Box-Cox) parameter
  ##
  if(!fix.lambda) {
    if(round(lambda, dig=4) == 1) {
      log.jacobian.max <- 0
      temp.list$z <- data
    }
    else {
      if(any(data^(lambda - 1) <= 0))
        log.jacobian.max <- log(prod(data^(lambda - 1)))
      else log.jacobian.max <- sum(log(data^(lambda - 1)))
      temp.list$z <- ((data^lambda)-1)/lambda
    }
  }
  else{
    log.jacobian.max <- temp.list$log.jacobian
  }
  ##
  ## Transforming coords for estimated anisotropy (if the case)
  ##
  if(fix.anR & fix.anA){remove(".likGRF.dists.vec", pos=1)}
  else{
    if(anR != 1 | anA != 0)
      coords <- coords.aniso(coords, aniso.pars=c(anA, anR))
    .likGRF.dists.vec <- as.vector(dist(coords))
    range.dist <- range(.likGRF.dists.vec)
    max.dist <- max(range.dist)
    min.dist <- min(range.dist)
    remove(".likGRF.dists.vec", pos=1)
  }      
  ##
  ## Computing estimated beta and tausq/sigmasq (if the case)
  ##
  if((phi < 1e-12))
    siv <- diag(x=1/sqrt((1+tausq)), n)
  else{
    if(check.sigmasq){
      if(sigmasq < 1e-12){
        if(!fix.nugget)
          siv <- diag(x=1/sqrt((1+tausq)), n)
        else
          siv <- diag(x=1/sqrt((tausq)), n)          
      }
      else
        siv <- varcov.spatial(coords = coords, cov.model = cov.model,
                              kappa = kappa,
                              nugget = tausq, cov.pars = c(1, phi),
                              func.inv = temp.list$func.inv, inv=TRUE,
                              sqrt.inv = TRUE, det = FALSE)$sqrt.inverse
    }
    else
      siv <- varcov.spatial(coords = coords, cov.model = cov.model,
                            kappa = kappa,
                            nugget = tausq, cov.pars = c(1, phi),
                            func.inv = temp.list$func.inv, inv=TRUE,
                            sqrt.inv = TRUE, det = FALSE)$sqrt.inverse
  }
  sivx <- crossprod(siv, temp.list$xmat)
  xivx <- crossprod(sivx)
  sivy <- crossprod(siv, temp.list$z)
  xivy <- crossprod(sivx, sivy)
  betahat <- solve(xivx, xivy)
  res <- as.vector(temp.list$z - temp.list$xmat %*% betahat)
  if(!fix.nugget | (nugget == 0)){
    res <- as.vector(temp.list$z - temp.list$xmat %*% betahat)
    ssres <- as.vector(crossprod(crossprod(siv,res)))
    if(method == "ML")
      sigmasq <- ssres/n
    else
      sigmasq <- ssres/(n - beta.size)
  }
  if(fix.nugget){
    if(nugget > 0)
      tausq <- nugget
  }
  else tausq <- tausq * sigmasq
  betahat.var <- solve(xivx)
  if(sigmasq > 1e-12) betahat.var <- sigmasq * betahat.var
#  if(!fix.nugget & phi < 1e-16){
#    tausq <- sigmasq + tausq
#    sigmasq <- 0
#  }
  ##
  ## Preparing output
  ##
  lik.results <- list(cov.model = cov.model,
                      nugget = tausq,
                      cov.pars=c(sigmasq, phi),
                      kappa = kappa,
                      beta = as.vector(betahat),
                      beta.var = betahat.var,
                      lambda = lambda,
                      aniso.pars = c(angle = anA, ratio = anR),
                      method = method,
                      loglik = loglik.max,
                      npars = npars,
                      AIC = (loglik.max - npars),
                      BIC = (loglik.max - 0.5 * log(n) * npars),
                      info = lik.optim,
                      max.dist = max.dist,
                      trend.matrix= temp.list$xmat,
                      log.jacobian = log.jacobian.max)
  ##
  ## Likelihood results for the model without spatial correlation
  ##
  if(nospatial){
    if(fix.lambda){
      beta.ns <- solve(crossprod(temp.list$xmat), crossprod(temp.list$xmat, temp.list$z))
      ss.ns <- sum((as.vector(temp.list$z - temp.list$xmat %*% beta.ns))^2)
      if(method == "ML"){
        nugget.ns <- ss.ns/n
        loglik.ns <- (n/2)*((-log(2*pi)) - log(nugget.ns) - 1) + temp.list$log.jacobian
      }
      if(method == "RML"){
        nugget.ns <- ss.ns/(n-beta.size)
        loglik.ns <- ((n-beta.size)/2)*((-log(2*pi)) - log(nugget.ns) -1) + temp.list$log.jacobian
      }
      npars.ns <- beta.size + 1 + fix.lambda
      lambda.ns <- lambda
    }
    else{
      bc.list <- list(n = n, beta.size = beta.size,
                      data = data, xmat = temp.list$xmat,
                      method = method)
      if(is.R())
        lik.lambda.ns <- optim(par=1, fn = boxcox.ns, method="L-BFGS-B",
                               lower= -3, upper= 3, bc.list=bc.list)
      else
        lik.lambda.ns <- nlminb(par=1, fn = boxcox.ns,
                                lower=-3, upper=3, data=data,
                                bc.list = bc.list)
      bc.list <- NULL
      lambda.ns <- lik.lambda.ns$par
      tdata.ns <- ((data^lambda.ns)-1)/lambda.ns
      beta.ns <- solve(crossprod(temp.list$xmat),crossprod(temp.list$xmat,tdata.ns))
      ss.ns <- sum((as.vector(tdata.ns - temp.list$xmat %*% beta.ns))^2)
      if(is.R())
        value.min.ns <- lik.lambda.ns$value
      else
        value.min.ns <- lik.lambda.ns$objective
      if(method == "ML"){
        loglik.ns <- (- value.min.ns)+ (n/2)*((-log(2*pi)) + log(n) - 1)
        nugget.ns <- ss.ns/n
      }
      if(method == "RML"){
        nugget.ns <- ss.ns/(n-beta.size)
        loglik.ns <- (- value.min.ns)+ ((n-beta.size)/2)*((-log(2*pi)) +
                                                          log(n-beta.size) - 1)
      }      
      npars.ns <- beta.size + 1 + fix.lambda
    }
    lik.results$nospatial <- list(beta.ns = beta.ns, variance.ns = nugget.ns,
                                  loglik.ns = loglik.ns, npars.ns = npars.ns,
                                  lambda.ns = lambda.ns)
  }
  ##
  ## Computing residuals and predicted values
  ## (isolated components of the model)
  ##
  if (components) {
    if(!fix.anR & !fix.anA)
      if(anR != 1 | anA != 0)
        coords <- coords.aniso(coords, aniso.pars=c(anA, anR))
    trend.comp <- temp.list$z - res
    invcov <- varcov.spatial(coords = coords, cov.model = cov.model, 
                             kappa = kappa, nugget = tausq,
                             func.inv = temp.list$func.inv,
                             cov.pars = c(sigmasq, phi), inv.TRUE)$inverse 
    covmat.signal <- varcov.spatial(coords = coords, cov.model = cov.model, 
                                    kappa = kappa, nugget = 0,
                                    cov.pars = c(sigmasq, phi))$varcov
    spatial.comp <- as.vector(covmat.signal %*% invcov %*% res)
    predict.comp <- trend.comp + spatial.comp
    residual.comp <- as.vector(temp.list$z - predict.comp)
    residual.std <- as.vector(invcov %*% residual.comp)
    residual.trend.std <- as.vector(invcov %*% res)
    s2.random <- (crossprod(res,invcov) %*% res)/(n - beta.size)
    s2 <- (crossprod(residual.comp,invcov) %*% residual.comp)/(n - beta.size)
  }
  if (length(lik.results$beta.var) == 1)
    lik.results$beta.var <- as.vector(lik.results$beta.var)
  if (length(lik.results$beta) > 1){
    if(inherits(trend, "formula"))
      beta.names <- c("intercept", paste("covar", 1:(ncol(temp.list$xmat)-1), sep = ""))
    else
      if (trend == "1st")
        beta.names <- c("1", "x", "y")
      else
        if (trend == "2nd")
          beta.names <- c("1", "x", "y", "x2", "xy", "y2")
    names(lik.results$beta) <- beta.names
  }
  if (components) {
    lik.results$model.components <- cbind(trend = trend.comp, spatial = spatial.comp, residual = residual.comp)
    lik.results$s2 <- s2
    lik.results$s2.random <- s2.trend
  }
  ##
  ## Printing some messages and finishing
  ##
  lik.results$call <- call
  class(lik.results) <- "variomodel"
  if(messages.screen == TRUE){
    cat("likfit: estimated model parameters are:\n")
    cat(paste("covariance model:", cov.model))
    if(cov.model == "matern" | cov.model == "powered.exponential" | 
       cov.model == "cauchy" | cov.model == "gneiting.matern")
      cat(paste(" with kappa =", kappa))
    if(!is.null(kappa))
      if(cov.model == "matern" & kappa == 0.5)
        cat(" (exponential)")
    cat("\n")
    print(c(nugget=lik.results$nugget, sill=lik.results$cov.pars[1], range=lik.results$cov.pars[2]))
    if (fix.lambda == FALSE)
      cat(paste("Box-Cox transformation parameter:", round(lik.results$lambda, dig=4),"\n"))
    if((lik.results$cov.pars[1] < (0.01 * (lik.results$nugget + lik.results$cov.pars[1])))& lik.results$cov.pars[2] > 0)
      cat("\nWARNING: estimated sill is less than 1 hundredth of the total variance. Consider re-examine the model excluding spatial dependence\n" )      
    if((lik.results$cov.pars[2] > (10 * max.dist)) & lik.results$cov.pars[1] > 0 )
      cat("\nWARNING: estimated range is more than 10 times bigger than the biggest distance between two points. Consider re-examine the model:\n 1) excluding spatial dependence if estimated sill is too low and/or \n 2) taking trends (covariates) into account\n" ) 
    if(((lik.results$cov.pars[2] < (0.1 * min.dist)) & (lik.results$cov.pars[1] > 0)) & lik.results$cov.pars[2] > 0)
      cat("\nWARNING: estimated range is less than 1 tenth of the minimum distance between two points. Consider re-examine the model excluding spatial dependence\n" ) 
  }
  ##
  ## Assigning classes
  ##
  if(is.R())
    class(lik.results) <- c("variomodel", "likGRF")
  else{
    if(version$major <= 4)
    class(lik.results) <- c("variomodel", "likGRF")
    else attr(kb.results, "class") <- setOldClass(c("variomodel", "likGRF"))
  }
  ##
  return(lik.results)
}

"negloglik.GRF" <-
  function(pars, fp, ip, temp.list)
### pars : values for the parameters to be estimated
  ##sequence is c(phi, tausq, kappa, lambda, anR, anA, sigmasq)
### fixed pars: parameters considered fixed
### ind.pars : list indicating which are fixed and which are to be estimated
  ##
  ##
  ## Warning:
  ##  if fix.nugget = TRUE and nugget > 0 ,
  ## sigmasq should be passed and fp$nugget is the value of the nugget
  ## otherwise the RELATIVE nugget should be passed
{
  ## Obligatory parameter:
  phi <- pars[1]
  ## Others
  if(ip$f.tausq){
    if(fp$tausq > 0){
      npars.min <- length(pars)
      sigmasq <- pars[npars.min]
    }
    else sigmasq <- 1
  }
  else sigmasq <- 1
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- fp$anA
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- pars[2]
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- pars[2]
    anA <- fp$anA
  }
  if(ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- pars[2]
    anA <- pars[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    anR <- fp$anR
    anA <- fp$anA
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    anR <- fp$anR
    anA <- pars[3]
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    anR <- pars[3]
    anA <- fp$anA
  }
  if(ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- fp$kappa
    lambda <- pars[2]
    anR <- pars[3]
    anA <- pars[4]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- fp$anA
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- pars[3]
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    anR <- pars[3]
    anA <- fp$anA
  }
  if(ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- fp$lambda
    anR <- pars[3]
    anA <- pars[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    anR <- fp$anR
    anA <- fp$anA
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    anR <- fp$anR
    anA <- pars[4]
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    anR <- pars[4]
    anA <- fp$anA
  }
  if(ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- fp$tausq
    kappa <- pars[2]
    lambda <- pars[3]
    anR <- pars[4]
    anA <- pars[5]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- fp$anA
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- pars[3]
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- pars[3]
    anA <- fp$anA
  }
  if(!ip$f.tausq & ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- fp$lambda
    anR <- pars[3]
    anA <- pars[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    anR <- fp$anR
    anA <- fp$anA
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    anR <- fp$anR
    anA <- pars[4]
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    anR <- pars[4]
    anA <- fp$anA
  }
  if(!ip$f.tausq & ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- fp$kappa
    lambda <- pars[3]
    anR <- pars[4]
    anA <- pars[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- fp$anA
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    anR <- fp$anR
    anA <- pars[4]
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    anR <- pars[4]
    anA <- fp$anA
  }
  if(!ip$f.tausq & !ip$f.kappa & ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- fp$lambda
    anR <- pars[4]
    anA <- pars[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    anR <- fp$anR
    anA <- fp$anA
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    anR <- fp$anR
    anA <- pars[5]
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    anR <- pars[5]
    anA <- fp$anA
  }
  if(!ip$f.tausq & !ip$f.kappa & !ip$f.lambda & !ip$f.anR & !ip$f.anA){
    tausq <- pars[2]
    kappa <- pars[3]
    lambda <- pars[4]
    anR <- pars[5]
    anA <- pars[6]
  }
  ##
  ##
  ##
  if(round(1e+16*(tausq+sigmasq)) == 0)
    return(1e+64)
  else{
    n <- temp.list$n
    p <- temp.list$beta.size
    log.jacobian <- temp.list$log.jacobian
    if(!ip$f.anR | !ip$f.anA){
      coords.c <- coords.aniso(temp.list$coords, aniso.pars=c(anA, anR))
      .likGRF.dists.vec <<- as.vector(dist(coords.c))
    }
    if(!ip$f.lambda){
      if(round(lambda, dig=4) == 1) {
        log.jacobian <- 0
      }
      else {
        if(any(temp.list$z <= 0))
          stop("Transformation not allowed for zero or negative data")
        data <- temp.list$z^(lambda - 1)
        if(any(data <= 0)) log.jacobian <- log(prod(data))
        else log.jacobian <- sum(log(data))
        data <- NULL
      }
      if(round(lambda, dig=4) == 0)
        data <- log(temp.list$z)
      else data <- ((temp.list$z^lambda) - 1)/lambda
    }
    else data <- temp.list$z
    ##
    ## Likelihood for Independent observations 
    ## (arbitrary criteria used here:
    ##      (phi < 1-e10) or (sigmasq < 1-e10)  ==> independence
    ##
    if((phi < 1e-12) | (sigmasq < 1e-12)){
      if(ip$f.tausq)
        iv <- list(sqrt.inverse = diag(x=1/sqrt((tausq+sigmasq)), n),
                   log.det.to.half = (n/2) * log(tausq+sigmasq))
      else
        iv <- list(sqrt.inverse = diag(x=1/sqrt((1+tausq)), n),
                   log.det.to.half = (n/2) * log(1+tausq))
    }
    else{
      iv <- varcov.spatial(dists.lowertri = .likGRF.dists.vec,
                           cov.model = temp.list$cov.model, kappa=kappa,
                           nugget = tausq, cov.pars=c(sigmasq, phi),
                           func.inv = temp.list$func.inv,
                           sqrt.inv = TRUE, det = TRUE)
    }
    sivx <- crossprod(iv$sqrt.inverse, temp.list$xmat)
    xivx <- crossprod(sivx)
    sivy <- crossprod(iv$sqrt.inverse, data)
    xivy <- crossprod(sivx, sivy)  
    betahat <- solve(xivx, xivy)
    res <- data - temp.list$xmat %*% betahat
    ssres <- as.vector(crossprod(crossprod(iv$sqrt.inverse,res)))
    if(temp.list$method == "ML"){
      if(ip$f.tausq & (tausq > 0))
        negloglik <- iv$log.det.to.half +  0.5 * ssres - log.jacobian
      else
        negloglik <- (n/2) * log(ssres) +  iv$log.det.to.half - log.jacobian
      return(negloglik)
    }
    if(temp.list$method == "RML"){
      if(length(as.vector(xivx)) == 1) {
        choldet <- 0.5 * log(xivx)
      }
      else {
        chol.xivx <- chol(xivx)
        choldet <- sum(log(diag(chol.xivx)))
      }
      if(ip$f.tausq & (tausq > 0))
        negloglik <- iv$log.det.to.half +  0.5 * ssres + choldet - log.jacobian
      else
        negloglik <- ((n-p)/2) * log(ssres) +  iv$log.det.to.half + choldet - log.jacobian
      if(negloglik > 1e64) negloglik <- 1e64
      return(negloglik)
    }
  }
}
 
"boxcox.ns" <- function(lambda, bc.list)
{
  data <- bc.list$data
  n <- bc.list$n
  ##
  if(round(lambda, dig=4) == 1) {
    log.jacobian <- 0
    y <- data
  }
  else {
    if(any(data <= 0))
      stop("Transformation option not allowed when there are zeros or negative data")
    Jdata <- data^(lambda - 1)
    if(any(Jdata <= 0))
      log.jacobian <- log(prod(Jdata))
    else log.jacobian <- sum(log(Jdata))
    Jdata <- NULL
    if(round(lambda, dig=4) == 0)
      y <- log(data)
    else y <- ((data^lambda) - 1)/lambda
  }
  beta.ns <- solve(crossprod(bc.list$xmat), crossprod(bc.list$xmat, y))
  ss.ns <- sum((as.vector(y) - as.vector(bc.list$xmat %*% beta.ns))^2)
  if(bc.list$method == "ML")
    neglik <- (n/2) * log(ss.ns) - log.jacobian
  if(bc.list$method == "RML")
    neglik <- ((n-bc.list$beta.size)/2) * log(ss.ns) - log.jacobian
  ##
  return(as.vector(neglik))
}

"likfit.limits" <-
  function(phi = list(lower=0, upper=+Inf),
           sigmasq = list(lower=0, upper=+Inf),
           tausq = list(lower=0, upper=+Inf),
           kappa = list(lower=0, upper=+Inf),
           lambda = list(lower=-3, upper=3),
           anR = list(lower=0, upper=+Inf),
           anA = list(lower=0, upper=2*pi)
           )
  {
    return(list(phi=phi, sigmasq=sigmasq,
                tausq = tausq, kappa = kappa,
                lambda = lambda, anR = anR, anA = anA))
  }
